from __future__ import annotations
import json

import re
from dataclasses import dataclass
from typing import Any, List, Optional, Sequence, Tuple, Type, Union

import requests
import bs4
from bs4 import BeautifulSoup
from bs4.element import PageElement

from .utils import tags_to_soup, render_list, get_heading_siblings_on_level


class WiktionaryPageNotFound(LookupError):
    pass


class LanguageEntryNotFound(LookupError):
    pass


@dataclass
class WordInfo:
    """
    Information about a word, including its language, part of speech, and definitions.
    """

    LANGUAGE = 'Spanish'
    entry: LanguageEntry

    @classmethod
    def from_search(cls, word: str, *, revision: Optional[int] = None):
        """
        Fetch information about a word from Wiktionary and return a WordInfo object.

        Parameters:
            word (str): The word to search for.
            revision (Optional[int]): The revision number of the Wiktionary page to use. If not
                provided, the latest revision will be used.
        """
        entry = WiktionaryPage(word, revision).get_entry(cls.LANGUAGE)
        return cls(entry=entry)

    @property
    def word(self) -> str:
        """The word represented by this WordInfo object, as a string"""
        return self.entry.page.word

    @property
    def definition_strings(self) -> List[str]:
        """Definitions of the word as a list of strings"""
        return [d.text for d in self.entry.definitions]

    @property
    def sections_with_definitions(self) -> List[Section]:
        """List of Section objects that contain definitions"""
        return self.entry.get_sections_with_definitions()

    def definition_output(self) -> str:
        """
        Human-readable multiline string with all definitions listed under its corresponding part of
        speech
        """
        outputs = []
        for section in self.entry.sections:
            if section.has_definitions():
                outputs.append(
                    f'{_render_section_lead(section)}\n'
                    f'{render_list(d.to_str() for d in section.definitions)}'
                )
        return '\n\n'.join(outputs)

    def compact_definition_output(self) -> str:
        """
        Human-readable multiline string with word and all definitions listed
        one after one another
        """
        definitions_with_bullet = [f'- {d.to_str()}' for d in self.entry.definitions]
        lines = [f'[bold yellow]{self.word}[/]'] + definitions_with_bullet
        return '\n'.join(lines)

    def json_output(self) -> str:
        """
        Return a JSON string representation of all information related to this WordInfo object.
        """
        return json.dumps(self.to_dict(), indent=2)

    def to_dict(self) -> dict:
        """
        Return a dict of all information related to this WordInfo object.
        """
        return dict(
            word=self.word,
            language=self.LANGUAGE,
            definition_sections=[d.to_dict() for d in self.sections_with_definitions],
        )


def _render_section_lead(ss: Section) -> str:
    """
    Render the lead for a section in a string format, including formatting tags for the `rich`
    package used to print to command line.

    The lead for a section consists of the part of speech, the word, the gender (if applicable),
    and any lead extras.

    Parameters:
        ss (Section): The Section object to render the lead for.

    Returns:
        str: The rendered lead for the given section.

    Example:
        >>> section = WiktionaryPage('olvidar').get_entry('Spanish').get_section('Verb')
        >>> _render_section_lead(section)
        '[italic]Verb:[/] [bold yellow]olvidar[/] ([italic]first-person singular present[/] \
[yellow]olvido[/], [italic]first-person singular preterite[/] [yellow]olvidé[/], \
[italic]past participle[/] [yellow]olvidado[/])'
    """
    parts = [f'[italic]{ss.part_of_speech}:[/]', f'[bold yellow]{ss.word}[/]']
    if ss.gender:
        parts.append(ss.gender)
    if ss.lead_extras:
        lead_extra_strings = _render_section_lead_extras(ss.lead_extras)
        parts.append(f"({', '.join(lead_extra_strings)})")
    return ' '.join(parts)


def _render_section_lead_extras(lead_extras: List[dict]) -> List[str]:
    """
    Render the lead extras for a section in a string format, including formatting tags for the
    `rich` package used to print to command line.

    Lead extras are attributes that are listed after the part of speech and word in the lead for
    a section.

    Parameters:
        lead_extras (List[dict]): A list of dictionaries representing the lead extras for a section.

    Returns:
        List[str]: The rendered lead extras for the given section.

    Example:
        >>> _render_section_lead_extras([
        ...     dict(attribute='first-person singular present', value='olvido'),
        ...     dict(attribute='first-person singular preterite', value='olvidé'),
        ...     dict(attribute='past participle', value='olvidado'),
        ... ])
        ['[italic]first-person singular present[/] [yellow]olvido[/]', \
'[italic]first-person singular preterite[/] [yellow]olvidé[/]', \
'[italic]past participle[/] [yellow]olvidado[/]']
    """
    lead_extra_strings = []
    for le in lead_extras:
        if 'value' in le:
            lead_extra_strings.append(
                f'[italic]{le["attribute"]}[/] [yellow]{le["value"]}[/]'
            )
        else:
            lead_extra_strings.append(f'[italic]{le["attribute"]}[/]')
    return lead_extra_strings


class WiktionaryPage:
    """
    Represents a page on Wiktionary, with the page content parsed into a BeautifulSoup object in
    the `soup` attribute.

    Use `get_entry()` to extract a particular LanguageEntry object from the page.
    """

    word: str
    revision: Optional[int] = None

    def __init__(self, word: str, revision: Optional[int] = None):
        """
        Initialize a WiktionaryPage object with a word and an optional revision number.

        Parameters:
            word (str): The word that this Wiktionary page is for.
            revision (Optional[int]): The revision number of the Wiktionary page to use. This is
                the `oldid` parameter in the Wiktionary page URL. If not provided, the latest
                revision will be used.
        """
        self.word = word
        self.revision = revision
        self.soup = BeautifulSoup(
            markup=self.get_page_html(word, revision), features='html.parser'
        )

    def __repr__(self):
        if self.revision is None:
            return f'{self.__class__.__name__}({self.word!r})'
        else:
            return (
                f'{self.__class__.__name__}({self.word!r}, revision={self.revision!r})'
            )

    @staticmethod
    def get_page_html(word, revision=None):
        """
        Retrieve the HTML content of the Wiktionary page for the given word and revision.

        Parameters:
            word (str): The word to retrieve the Wiktionary page for.
            revision (Optional[int]): The revision number of the Wiktionary page to use. If not
                provided, the latest revision will be used.

        Returns:
            str: The HTML content of the Wiktionary page.

        Raises:
            WiktionaryPageNotFound: If the Wiktionary page for the given word cannot be found.
        """
        if revision is None:
            url = f'https://en.wiktionary.org/wiki/{word}'
        else:
            url = f'https://en.wiktionary.org/w/index.php?title={word}&oldid={revision}'
        content = request_url_text(url)
        if 'Wiktionary does not yet have an entry for' in content:
            raise WiktionaryPageNotFound('No Wiktionary page found')
        return content

    def __eq__(self, other) -> bool:
        """
        Check if this WiktionaryPage object is equal to another object.

        Two WiktionaryPage objects are considered equal if they have the same word, revision
        number, and the `soup` BeautifulSoup objects are equal.
        """
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.word == other.word
            and self.revision == other.revision
            and self.soup == other.soup
        )

    def get_spanish_entry(self) -> LanguageEntry:
        return self.get_entry(language='Spanish')

    def get_entry(self, language: str) -> LanguageEntry:
        """
        Retrieve the language entry for the given language on this Wiktionary page.

        Parameters:
            language (str): The language to retrieve the entry for.

        Returns:
            LanguageEntry: The language entry for the given language on this Wiktionary page.

        Raises:
            LanguageEntryNotFound: If a language entry for the given language cannot be found on
                this Wiktionary page.
        """
        return LanguageEntry(
            soup=_extract_language_entry_soup(self.soup, language=language), page=self
        )


def _extract_language_entry_soup(
    page_soup: BeautifulSoup, language: str
) -> BeautifulSoup:
    """
    Get a new BeautifulSoup object that only has the tags from the entry that matches
    `language`.
    """
    tags = _language_entry_tags(page_soup, language)
    return tags_to_soup(tags)


def _language_entry_tags(page_soup: BeautifulSoup, language: str) -> List[PageElement]:
    """
    Get a list of all BeautifulSoup elements since under the heading that matches `language`.
    """
    start_tag = _entry_start_tag(page_soup, language)
    return get_heading_siblings_on_level(start_tag)


def _entry_start_tag(page_soup: BeautifulSoup, language: str) -> bs4.Tag:
    """
    Find the first `h2` tag in the given BeautifulSoup object representing a Wiktionary page. This
    is used to locate the beginning of the language entry for the word being searched.
    """
    id_tag: bs4.Tag = page_soup.find(id=language)
    if id_tag is None:
        raise LanguageEntryNotFound(f'No {language} entry found from page')
    start_tag = id_tag.parent
    assert start_tag.name == 'h2'
    return start_tag


class LanguageEntry:
    """
    A class representing the information for a single language entry on a Wiktionary page.

    A LanguageEntry object should be generated from a BeautifulSoup object that only contains the
    parsed HTML for one language (not the full page). `WiktionaryPage.get_entry()` does this with
    the `_extract_language_entry_soup()` helper.

    You can get the section through the WiktionaryPage like this:
    >>> entry = WiktionaryPage('empleado').get_entry('Spanish')
    >>> isinstance(entry, LanguageEntry)
    True
    """

    # TODO clear up the hierarchy and inheritance between LanguageEntry and Section.

    def __init__(self, soup: BeautifulSoup, page: WiktionaryPage):
        # self.title = title
        self.page = page
        self.soup = soup

    def __repr__(self):
        return f'<{self.page!r} → {self.title!r}>'

    @property
    def title(self) -> str:
        return self.soup.find(class_='mw-headline').text

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.page == other.page and self.soup == other.soup

    @property
    def sections(self) -> List[Section]:
        """
        Get a list of Section objects inside this language entry.

        Sections correspond to the subheadings under one language entry on the Wiktionary page.
        Some, but not all, of the sections are parts of speech: E.g. the page for 'empleado' has
        sections 'Etymology', 'Pronunciation', 'Adjective', 'Noun', 'Participle', Further reading'.
        """
        level = 'h3'
        subheadings = self.soup.find_all(level)
        tag_sets = [get_heading_siblings_on_level(sh) for sh in subheadings]
        subsoups = [tags_to_soup(tags) for tags in tag_sets]
        return [Section(parent=self, soup=subsoup) for subsoup in subsoups]

    def get_section(self, title: str) -> Section:
        """
        Get a specific section by its title (exact string match).

        Example:
            >>> entry = WiktionaryPage('empleado').get_entry('Spanish')
            >>> section = entry.get_section('Noun')
            >>> isinstance(section, Section)
            True
        """
        for ss in self.sections:
            if ss.title == title:
                return ss
        else:
            raise KeyError(f'No section with title: {title}')

    def get_sections_with_definitions(self) -> List[Section]:
        """Get a list of all sections that contain any definitions."""
        return [sub for sub in self.sections if sub.has_definitions()]

    @property
    def definitions(self) -> List[Definition]:
        """Get a list of all definitions contained in sections of this LanguageEntry"""
        definitions_ = []
        for sub in self.get_sections_with_definitions():
            definitions_.extend(sub.definitions)
        return definitions_


_EMPTY_TAG = bs4.Tag(name='empty')
ConjugationTableDiv = Type[bs4.Tag]
Conjugation = dict


class Section(LanguageEntry):
    def __init__(self, parent: LanguageEntry, soup: BeautifulSoup):
        self.parent = parent
        if not isinstance(parent, LanguageEntry) or isinstance(parent, Section):
            raise TypeError('parent has to be LanguageEntry and cannot be Section')
        self.soup = soup

    def __repr__(self):
        return f'<{self.parent.page} → {self.parent.title!r} → {self.title!r}>'

    def to_dict(self) -> dict[str, Any]:
        if self.has_definitions():
            D: dict[str, Any] = dict(
                part_of_speech=self.part_of_speech,
                word=self.word,
                extras=self.lead_extras,
                definitions=[d.to_dict() for d in self.definitions],
            )
            if self.conjugation is not None:
                D['conjugation'] = self.conjugation
            return D
        else:
            return dict()

    @property
    def part_of_speech(self) -> str:
        return self.title

    @property
    def word(self) -> str:
        return self._word_tag.text

    @property
    def _word_tag(self) -> bs4.Tag:
        return self._lead_p.find(class_='headword') or _EMPTY_TAG

    @property
    def _lead_p(self) -> bs4.Tag:
        return self.soup.p or _EMPTY_TAG

    @property
    def conjugation(self) -> Optional[Conjugation]:
        # TODO raise NotImplemented error if trying for languages other than Spanish
        table_root = self._conjugation_table_root()
        if table_root is None:
            return None

        return ConjugationTable(table_root).to_dict()

    def _conjugation_table_root(self) -> Optional[ConjugationTableDiv]:
        headings = self.soup.find_all('h4')
        candidate_table_headings = [h for h in headings if h.find(string='Conjugation')]
        try:
            table_heading = candidate_table_headings[0]
        except IndexError:
            return None
        return table_heading.find_next('div', class_='NavFrame')

    # TODO write a specific test
    @property
    def gender(self) -> Optional[str]:
        tag = self._lead_p.find(class_='gender')
        if tag is None:
            return None
        return tag.text

    @property
    def lead_extras(self) -> List[dict]:
        """
        Extract attributes from inside parentheses on the lead line (after the word).
        """
        word_tag = self._word_tag
        opening_parenthesis = (
            word_tag.find_next_sibling(string=re.compile(r'\(')) or _EMPTY_TAG
        )

        # take attributes from <i> tags and corresponding values from the following <b> tag
        # TODO this seems quite fragile
        L = []
        for attribute_tag in opening_parenthesis.find_next_siblings('i'):
            value_tag = attribute_tag.find_next_sibling('b')
            D = {'attribute': attribute_tag.text}
            if value_tag:  # only try to extract text if the value tab exists
                D['value'] = value_tag.text
            L.append(D)
        return L

    @property
    def definitions(self) -> List[Definition]:
        """
        Parse definitions from soup and return them as a list of strings.
        """
        return [
            Definition(
                text=self.definition_list_item_to_str(definition_list_item),
                extras=None,
                section=self,
            )
            for definition_list_item in self._definition_list_items()
        ]

    def has_definitions(self) -> bool:
        return len(self.definitions) > 0

    def _definition_list_items(self) -> List[bs4.Tag]:
        return self._definition_list_items_from_soup(self.soup)

    @staticmethod
    def _definition_list_items_from_soup(soup: BeautifulSoup) -> List[bs4.Tag]:
        """
        Extract <li> tags that contain definitions
        """
        dlis = []
        for ol in soup.find_all('ol', recursive=False):
            for child in ol.find_all('li', recursive=False):
                dlis.append(child)
        return dlis

    @staticmethod
    def definition_list_item_to_str(li: bs4.Tag) -> str:
        """
        Parse the contents of the given definition `li` tag.
        """
        res = []
        for e in li.children:
            if e.name not in ('dl', 'ul'):  # exclude nested stuff
                res.append(e.get_text())
        return ''.join(res).strip()


class ConjugationTable:

    _ROW_INDEX_INFINITIVE = 0
    _ROW_INDEX_GERUND = 1
    _ROW_SLICE_PAST_PARTICIPLE = slice(3, 5)
    _ROW_SLICE_INDICATIVE = slice(8, 13)
    _ROW_SLICE_SUBJUNCTIVE = slice(15, 19)
    _ROW_SLICE_IMPERATIVE = slice(21, 23)

    def __init__(self, root: ConjugationTableDiv):
        """
        Initialize from a "root" tag, a div that contains the conjugation table
        """
        self.root = root
        self.table = self.root.find('table')

    def to_dict(self):
        return {
            'infinitive': self._parse_simple(self._ROW_INDEX_INFINITIVE),
            'gerund': self._parse_simple(self._ROW_INDEX_GERUND),
            'past participle': self._parse_complex(
                self._ROW_SLICE_PAST_PARTICIPLE, header=('masculine', 'feminine')
            ),
            'indicative': self._parse_complex(
                self._ROW_SLICE_INDICATIVE,
                header=('s1', 's2', 's3', 'pl1', 'pl2', 'pl3'),
            ),
            'subjunctive': self._parse_complex(
                self._ROW_SLICE_SUBJUNCTIVE,
                header=('s1', 's2', 's3', 'pl1', 'pl2', 'pl3'),
            ),
            'imperative': self._parse_complex(
                self._ROW_SLICE_IMPERATIVE,
                header=('s1', 's2', 's3', 'pl1', 'pl2', 'pl3'),
            ),
        }

    def _parse_simple(self, row_index: int):
        tag = self.table.tbody.find_all('tr')[row_index].td
        return tag.get_text().strip()

    def _parse_complex(self, row_slice: slice, header: Sequence[str]):
        d = {}
        for tr in self.table.tbody.find_all('tr')[row_slice]:
            row_key, values_dict = self._parse_row(tr, header)
            d[row_key] = values_dict
        return d

    @staticmethod
    def _parse_row(tr: bs4.Tag, header: Sequence[str]) -> Tuple[str, dict]:
        row_key = tr.th.get_text().strip()
        value_tags = [td for td in tr.find_all('td')]
        values = [ConjugationTable._parse_value_tag(td) for td in value_tags]
        values_dict = {h: v for h, v in zip(header, values)}
        return row_key, values_dict

    @staticmethod
    def _parse_value_tag(td: bs4.Tag) -> Union[str, dict, None]:
        """
        Parse a <td> tag containing one conjugation.

        In the usual case, when the td contains one span, return the text inside it (or None if
        the text is empty). In the tuteo/voseo case, the td contains 2 spans; their texts are
        split in a dict.
        """
        spans = td.find_all('span')
        if len(spans) == 1:
            text = spans[0].get_text().strip() or None
            return text
        elif len(spans) > 1:
            return {
                'tú': spans[0].get_text().strip(),
                'vos': spans[1].get_text().strip(),
            }
        else:
            return None


@dataclass
class Definition:
    text: str
    extras: Optional[dict]  # we would put synonyms, antonyms, usage examples, etc. here
    section: Section

    def to_str(self) -> str:
        return self.text

    def to_dict(self) -> dict:
        return dict(text=self.text)


def request_url_text(url: str) -> str:
    return requests.get(url).text  # pragma: no cover



import json
from pathlib import Path
from textwrap import dedent

import pytest
from bs4 import BeautifulSoup
from pytest_mock import MockerFixture

import palabras.core
import palabras.cli
from palabras.core import Section, WiktionaryPage, WordInfo
from palabras.utils import get_siblings_until, get_heading_siblings_on_level


MOCK_CACHE_FILE_PATH = Path(__file__).parent / '../data/mock_cache.json'

EXPECTED_CONJUGATION_OLVIDAR = {
    'infinitive': 'olvidar',
    'gerund': 'olvidando',
    'past participle': {
        'singular': {
            'masculine': 'olvidado',
            'feminine': 'olvidada'
        },
        'plural': {
            'masculine': 'olvidados',
            'feminine': 'olvidadas'
        }
    },
    'indicative': {
        'present': {
            's1': 'olvido',
            's2': {
                'tú': 'olvidas',
                'vos': 'olvidás'
            },
            's3': 'olvida',
            'pl1': 'olvidamos',
            'pl2': 'olvidáis',
            'pl3': 'olvidan'
        },
        'imperfect': {
            's1': 'olvidaba',
            's2': 'olvidabas',
            's3': 'olvidaba',
            'pl1': 'olvidábamos',
            'pl2': 'olvidabais',
            'pl3': 'olvidaban'
        },
        'preterite': {
            's1': 'olvidé',
            's2': 'olvidaste',
            's3': 'olvidó',
            'pl1': 'olvidamos',
            'pl2': 'olvidasteis',
            'pl3': 'olvidaron'
        },
        'future': {
            's1': 'olvidaré',
            's2': 'olvidarás',
            's3': 'olvidará',
            'pl1': 'olvidaremos',
            'pl2': 'olvidaréis',
            'pl3': 'olvidarán'
        },
        'conditional': {
            's1': 'olvidaría',
            's2': 'olvidarías',
            's3': 'olvidaría',
            'pl1': 'olvidaríamos',
            'pl2': 'olvidaríais',
            'pl3': 'olvidarían'
        }
    },
    'subjunctive': {
        'present': {
            's1': 'olvide',
            's2': {
                'tú': 'olvides',
                'vos': 'olvidés'
            },
            's3': 'olvide',
            'pl1': 'olvidemos',
            'pl2': 'olvidéis',
            'pl3': 'olviden'
        },
        'imperfect(ra)': {
            's1': 'olvidara',
            's2': 'olvidaras',
            's3': 'olvidara',
            'pl1': 'olvidáramos',
            'pl2': 'olvidarais',
            'pl3': 'olvidaran'
        },
        'imperfect(se)': {
            's1': 'olvidase',
            's2': 'olvidases',
            's3': 'olvidase',
            'pl1': 'olvidásemos',
            'pl2': 'olvidaseis',
            'pl3': 'olvidasen'
        },
        'future1': {
            's1': 'olvidare',
            's2': 'olvidares',
            's3': 'olvidare',
            'pl1': 'olvidáremos',
            'pl2': 'olvidareis',
            'pl3': 'olvidaren'
        }
    },
    'imperative': {
        'affirmative': {
            's1': None,
            's2': {
                'tú': 'olvida',
                'vos': 'olvidá'
            },
            's3': 'olvide',
            'pl1': 'olvidemos',
            'pl2': 'olvidad',
            'pl3': 'olviden'
        },
        'negative': {
            's1': None,
            's2': 'no olvides',
            's3': 'no olvide',
            'pl1': 'no olvidemos',
            'pl2': 'no olvidéis',
            'pl3': 'no olviden'
        }
    }
}
EXPECTED_DICT_OLVIDAR = dict(
    word='olvidar',
    language='Spanish',
    definition_sections=[
        dict(
            part_of_speech='Verb',
            word='olvidar',
            extras=[
                dict(attribute='first-person singular present', value='olvido'),
                dict(attribute='first-person singular preterite', value='olvidé'),
                dict(attribute='past participle', value='olvidado'),
            ],
            definitions=[
                dict(text='(transitive) to forget (be forgotten by)'),
                dict(text='(reflexive, intransitive) to forget, elude, escape'),
                dict(text='(with de, reflexive, intransitive) to forget, to leave behind')
            ],
            conjugation=EXPECTED_CONJUGATION_OLVIDAR
        )
    ]
)


@pytest.fixture()
def mocked_request_url_text(mocker: MockerFixture):
    """
    Mock palabras.core.request_url_text to read URL -> content (HTML) mappings from a pre-populated
    file instead of over the internet.
    """
    with open(MOCK_CACHE_FILE_PATH) as fp:
        mock_cache: dict = json.load(fp)['url_contents']
    mocker.patch('palabras.core.request_url_text', side_effect=lambda url: mock_cache[url])


@pytest.mark.parametrize('args', [
    ['-h'],
    ['-V'],
    ['olvidar'],
    ['olvidar', '-r', '66217360'],
    ['ene', '-r', '68912066'],
])
def test_main_exitcode_only(mocker, args):
    """
    Simulate calling the CLI with various sets of options and assert that the exitcode is 0.

    If a call is found to cause an unexpected error, throw those parameters here, fix the error,
    and we have a test to assert that it stays fixed.
    """
    mocker.patch('sys.argv', ['palabras'] + args)
    with pytest.raises(SystemExit) as e:
        # importing __main__ causes the script to be run
        from palabras import __main__  # noqa: F401
    assert e.value.code == 0


def test_word_info_from_search_return_type(mocked_request_url_text):
    word = 'despacito'
    wi = WordInfo.from_search(word)
    assert isinstance(wi, palabras.core.WordInfo)


def test_word_info_equals_but_not_is(mocked_request_url_text):
    word = 'despacito'
    wi1 = WordInfo.from_search(word)
    wi2 = WordInfo.from_search(word)
    assert wi1 == wi2
    assert wi1 is not wi2


def test_page_equalities(mocked_request_url_text):
    word = 'olvidar'
    wp_1 = WiktionaryPage(word=word)
    wp_2 = WiktionaryPage(word=word)

    assert wp_1 == wp_2
    assert wp_1 is not wp_2

    revision_a = 62345284
    revision_b = 66217360
    wp_with_revision = WiktionaryPage(word=word, revision=revision_a)
    wp_with_revision_same = WiktionaryPage(word=word, revision=revision_a)
    wp_with_revision_other = WiktionaryPage(word=word, revision=revision_b)

    assert wp_1 != wp_with_revision
    assert wp_with_revision == wp_with_revision_same
    assert wp_with_revision != wp_with_revision_other


def test_eqs_with_other_types(mocked_request_url_text):
    word = 'olvidar'
    page = WiktionaryPage(word)
    entry = page.get_entry('Spanish')
    wi = WordInfo(entry)
    assert page != 'foo'
    assert entry != 'foo'
    assert wi != 'foo'


def test_get_wiktionary_page_nonexistent(mocked_request_url_text):
    word = 'thispageaintexistent'
    with pytest.raises(palabras.core.WiktionaryPageNotFound):
        palabras.core.WiktionaryPage(word)


def test_get_wiktionary_page_contains_translation(mocked_request_url_text):
    word = 'culpar'
    translation = 'blame'
    page = palabras.core.WiktionaryPage(word)
    assert translation in str(page.soup)


def test_wiktionary_page_contains_portuguese_conjugation(mocked_request_url_text):
    word = 'culpar'
    expected_contains = 'culpou'  # Portuguese 3rd person preterite
    page = palabras.core.WiktionaryPage(word)
    assert expected_contains in str(page.soup)


def test_spanish_entry_type(mocked_request_url_text):
    word = 'culpar'
    result = WiktionaryPage(word).get_spanish_entry()
    assert isinstance(result, palabras.core.LanguageEntry)


def test_no_spanish_definition(mocked_request_url_text):
    word = 'kauppa'  # a word that has Wiktionary page but no Spanish definition
    with pytest.raises(palabras.core.LanguageEntryNotFound):
        WiktionaryPage(word).get_spanish_entry()


def test_spanish_entry_does_not_contain_portuguese(mocked_request_url_text):
    word = 'culpar'
    portuguese_conjugation = 'culpou'  # Portuguese 3rd person preterite
    spanish_entry = WiktionaryPage(word).get_spanish_entry()
    assert portuguese_conjugation not in str(spanish_entry.soup)


# TODO remove?
def test_definition_list_item_to_str(mocked_request_url_text):
    li = BeautifulSoup('''
    <li>parse <a href="foo">this</a><dl><dd><span>Whatever</span>...</dd></dl></li>
    ''', features='html.parser').li
    str_definition = palabras.core.Section.definition_list_item_to_str(li)
    assert str_definition == 'parse this'


def test_lookup_definition(mocked_request_url_text):
    word = 'culpar'
    wi = WordInfo.from_search(word)
    assert wi.definition_strings[0] == 'to blame'


def test_lookup_definition_complicated(mocked_request_url_text):
    word = 'empleado'  # this word has definitions for adjective, noun, and verb
    revision = 62175311
    wi = WordInfo.from_search(word, revision=revision)
    assert wi.definition_strings == [
        'employed',
        'employee',
        'Masculine singular past participle of emplear.'
    ]


def test_lookup_different_definitions_in_history(mocked_request_url_text):
    word = 'olvidar'
    revision_1 = 62345284
    revision_2 = 66217360
    expected_definitions_1 = [
        'to forget; to elude, escape (be forgotten by)',
        '(reflexive) to forget',
        '(reflexive) to leave behind'
    ]
    expected_definitions_2 = [
        '(transitive) to forget (be forgotten by)',
        '(reflexive, intransitive) to forget, elude, escape',
        '(with de, reflexive, intransitive) to forget, to leave behind'
    ]

    assert WordInfo.from_search(word, revision=revision_1).definition_strings \
        == expected_definitions_1
    assert WordInfo.from_search(word, revision=revision_2).definition_strings \
        == expected_definitions_2


def test_cli(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args = ['olvidar']
    palabras.cli.main(args)
    captured = capsys.readouterr()
    expected = dedent('''
        Verb: olvidar (first-person singular present olvido, first-person singular preterite olvidé, past participle olvidado)
        - (transitive) to forget (be forgotten by)
        - (reflexive, intransitive) to forget, elude, escape
        - (with de, reflexive, intransitive) to forget, to leave behind
    ''').lstrip()  # noqa: E501
    assert captured.out == expected


def test_cli_compact(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args = ['olvidar', '--compact']
    expected = dedent('''
        olvidar
        - (transitive) to forget (be forgotten by)
        - (reflexive, intransitive) to forget, elude, escape
        - (with de, reflexive, intransitive) to forget, to leave behind
    ''').lstrip()

    exitcode = palabras.cli.main(args)
    captured = capsys.readouterr()
    assert captured.out == expected
    assert exitcode == 0


def test_cli_json(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args = ['olvidar', '--json']
    exitcode = palabras.cli.main(args)
    captured = capsys.readouterr()
    assert json.loads(captured.out) == EXPECTED_DICT_OLVIDAR
    assert exitcode == 0


def test_cli_revision(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args1 = ['olvidar', '-r', '62345284']
    args2 = ['olvidar', '--revision', '62345284']
    args3 = ['olvidar', '--revision=62345284']

    expected = dedent('''
        Verb: olvidar (first-person singular present olvido, first-person singular preterite olvidé, past participle olvidado)
        - to forget; to elude, escape (be forgotten by)
        - (reflexive) to forget
        - (reflexive) to leave behind
    ''').lstrip()  # noqa: E501

    for args in args1, args2, args3:
        exitcode = palabras.cli.main(args)
        captured = capsys.readouterr()
        assert captured.out == expected
        assert exitcode == 0


def test_cli_ser(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args = ['ser']
    exitcode = palabras.cli.main(args)
    captured = capsys.readouterr()
    expected = dedent('''
        Verb: ser (first-person singular present soy, first-person singular preterite fui, past participle sido)
        - to be (essentially or identified as)
        - to be (in the passive voice sense)
        - to exist; to occur

        Noun: ser m (plural seres)
        - a being, organism
        - nature, essence
        - value, worth
    ''').lstrip()  # noqa: E501
    assert captured.out == expected
    assert exitcode == 0


def test_cli_ser_compact(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args = ['ser', '--compact']
    exitcode = palabras.cli.main(args)
    captured = capsys.readouterr()
    expected = dedent('''
        ser
        - to be (essentially or identified as)
        - to be (in the passive voice sense)
        - to exist; to occur
        - a being, organism
        - nature, essence
        - value, worth
    ''').lstrip()
    assert captured.out == expected
    assert exitcode == 0


def test_cli_nonexistent_page(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args = ['asdasdasd']
    exitcode = palabras.cli.main(args)
    captured = capsys.readouterr()
    expected = 'No Wiktionary page found\n'
    assert captured.out == expected
    assert exitcode == 1


def test_cli_no_spanish_entry(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args = ['moikka']
    exitcode = palabras.cli.main(args)
    captured = capsys.readouterr()
    expected = 'No Spanish entry found from page\n'
    assert captured.out == expected
    assert exitcode == 1


def test_main(capsys: pytest.CaptureFixture, mocker):
    """
    A little bit more opaque setup to test if we get expected output also by running __main__.py
    with mocked sys.argv
    """
    mocker.patch('sys.argv', ['palabras', 'hola'])  # mock cli arguments
    with pytest.raises(SystemExit) as e:
        # importing __main__ causes the script to be run
        from palabras import __main__  # noqa: F401
    assert e.value.code == 0  # check exitcode
    captured = capsys.readouterr()
    assert captured.out == dedent('''
        Interjection: ¡hola!
        - hello, hi
    ''').lstrip()


def test_get_next_siblings_until():
    markup = (
        '<h1>hello</h1>'
        '<p>foo</p>'
        '<h2>hello again</h2>'
        '<p>bar</p>'
        '<h1>hello 2</h1>'
        '<p>foo 2</p>'
        '<h2>hello again 2</h2>'
        '<p>bar 2</p>'
        '<h1>hello 3</h1>'
    )
    soup = BeautifulSoup(markup, features='html.parser')
    tag = soup.find('h1')
    assert len(get_siblings_until(tag, 'p')) == 1  # hello
    assert len(get_siblings_until(tag, 'h2')) == 2  # hello ... hello again
    assert len(get_siblings_until(tag, 'h1')) == 4  # hello ... hello 2
    assert len(get_siblings_until(tag, 'nope')) == 9  # all tags

    tag2 = soup.find_all('h1')[1]
    assert len(get_siblings_until(tag2, 'h1')) == 4  # hello 2 ... bar 2
    assert len(get_siblings_until(tag2, 'nope')) == 5  # hello 2 ... hello 3


def test_page_object(mocked_request_url_text):
    word = 'ser'
    page = WiktionaryPage(word)
    assert isinstance(page, WiktionaryPage)


def test_page_object_from_word_and_revision(mocked_request_url_text):
    word = 'empleado'
    revision = 62175311
    page = WiktionaryPage(word, revision)
    assert isinstance(page, WiktionaryPage)


def test_page_object_attributes(mocked_request_url_text):
    word = 'empleado'
    revision = 62175311
    page = WiktionaryPage(word, revision)
    assert page.word == word
    assert page.revision == revision


def test_sections_len_and_types(mocked_request_url_text):
    page = WiktionaryPage('empleado')
    entry = page.get_spanish_entry()
    sections = entry.sections
    assert len(sections) > 0  # this page has sections
    for s in sections:
        assert isinstance(s, Section)


def test_section_titles(mocked_request_url_text):
    page = WiktionaryPage('empleado', revision=68396093)
    entry = page.get_spanish_entry()
    titles = [s.title for s in entry.sections]
    assert titles == [
        'Etymology',
        'Pronunciation',
        'Adjective',
        'Noun',
        'Participle',
        'Further reading'
    ]


def test_get_specific_section(mocked_request_url_text):
    page = WiktionaryPage('empleado')
    entry = page.get_spanish_entry()
    section_adjective = entry.get_section('Adjective')
    assert isinstance(section_adjective, Section)
    assert section_adjective.title == 'Adjective'


def test_get_nonexistent_section(mocked_request_url_text):
    page = WiktionaryPage('empleado')
    entry = page.get_spanish_entry()
    with pytest.raises(KeyError, match='No section with title:'):
        entry.get_section('Nonexistent section')


def test_get_heading_siblings_on_level():
    soup = BeautifulSoup(
        '<h1>tag 1</h1>'
        '<h2>tag 2</h2>'
        '<h3>tag 3</h3>'
        '<p>tag 4</p>'
        '<h3>tag 5</h3>'
        '<h2>tag 6</h2>',
        features='html.parser'
    )
    element = soup.h2  # find the first h2
    # should find 2,3,4,5
    assert len(get_heading_siblings_on_level(element)) == 4

    element = soup.h3  # find the first h3
    # should find 3,4
    assert len(get_heading_siblings_on_level(element)) == 2


def test_get_siblings_on_level_error_on_unexpected_element():
    soup = BeautifulSoup(
        '<h1>one</h1>'
        '<li>get_siblings_on_level</li>'
        '<li>only works for headings</li>',
        features='html.parser'
    )
    element = soup.li
    with pytest.raises(ValueError):
        get_heading_siblings_on_level(element)


def test_page_repr(mocked_request_url_text):
    assert repr(WiktionaryPage('empleado')) \
        == "WiktionaryPage('empleado')"
    assert repr(WiktionaryPage('empleado', revision=62175311)) \
        == "WiktionaryPage('empleado', revision=62175311)"


def test_entry_repr(mocked_request_url_text):
    page = WiktionaryPage('empleado')
    entry = page.get_entry('Spanish')
    assert repr(entry) == "<WiktionaryPage('empleado') → 'Spanish'>"


def test_section_repr(mocked_request_url_text):
    page = WiktionaryPage('ser')
    entry = page.get_entry('Spanish')
    entry = entry.get_section('Verb')
    assert repr(entry) == "<WiktionaryPage('ser') → 'Spanish' → 'Verb'>"


def test_section_raises_on_invalid_parent():
    placeholder_soup = BeautifulSoup('<a>foo</a>', features='html.parser')
    page = WiktionaryPage('ser')

    # get a valid section from the page
    ok_section = page.get_entry('Spanish').sections[0]

    # parent cannot be a Section object
    with pytest.raises(TypeError):
        Section(parent=ok_section, soup=placeholder_soup)

    # parent cannot be None
    with pytest.raises(TypeError):
        Section(parent=None, soup=placeholder_soup)


def test_minimal_section_lead(mocker):
    mock_html = '''
    <h2><span id="EntryTitle"></span></h2>
    <h3><span class="mw-headline">SectionTitle</span></h3>
    <p>
        <span class="headword">headword</span>
        (<i>an attribute</i><b>a value</b>)
    </p>
    '''
    mocker.patch('palabras.core.WiktionaryPage.get_page_html', return_value=mock_html)

    page = WiktionaryPage('foo')
    entry = page.get_entry('EntryTitle')
    section = entry.get_section('SectionTitle')
    assert section.word == 'headword'
    assert section.lead_extras == [{'attribute': 'an attribute', 'value': 'a value'}]


def test_minimal_section_empty_lead(mocker):
    # mock to create page with a minimal HTML that doesn't have <p> under section
    mock_html = '''
    <h2><span id="EntryTitle"></span></h2>
        <h3><span class="mw-headline">SectionTitle</span></h3>
    '''
    mocker.patch('palabras.core.WiktionaryPage.get_page_html', return_value=mock_html)

    # create a section through a page object
    section = (
        WiktionaryPage('foo')
        .get_entry('EntryTitle')
        .get_section('SectionTitle')
    )

    assert section.word == ''
    assert section.lead_extras == []
    assert section.to_dict() == {}


def test_word_info_to_dict(mocked_request_url_text):
    wi = WordInfo.from_search('olvidar')
    assert wi.to_dict() == EXPECTED_DICT_OLVIDAR


def test_no_conjugation_for_noun():
    wi = WordInfo.from_search('palabra')
    assert wi.sections_with_definitions[0].conjugation is None


def test_verb_conjugation_is_dict():
    wi = WordInfo.from_search('olvidar')
    assert isinstance(wi.sections_with_definitions[0].conjugation, dict)


@pytest.mark.parametrize('keys, expected', (
    (['infinitive'], 'olvidar'),
    (['gerund'], 'olvidando'),
    (['past participle', 'plural', 'feminine'], 'olvidadas'),
    (['indicative', 'future', 'pl2'], 'olvidaréis'),
    (['subjunctive', 'present', 's2'], {'tú': 'olvides', 'vos': 'olvidés'}),  # tuteo/voseo case
    (['imperative', 'affirmative', 's1'], None),
))
def test_verb_conjugation_content(mocked_request_url_text, keys, expected):
    wi = WordInfo.from_search('olvidar')
    conjugation = wi.sections_with_definitions[0].conjugation

    # loop through keys and traverse the conjugation dict structure
    item = conjugation
    for key in keys:
        item = item[key]
    assert item == expected



from copy import copy
from typing import Container, List, Union, Sequence, Iterable
import bs4
from bs4 import BeautifulSoup
from bs4.element import PageElement


def tags_to_soup(tags: Sequence[bs4.Tag], *, features='html.parser') -> BeautifulSoup:
    """
    Given a list of tags, create a new BeautifulSoup object from those tags (copying tags to the
    new soup object in order)
    """
    soup = BeautifulSoup(features=features)
    for element in tags:
        soup.append(copy(element))
    return soup


def get_heading_siblings_on_level(element):
    """
    Return a list of sibling elements until the next occurrence of a heading on the same
    or higher level.
    """
    hierarchy = 'h1', 'h2', 'h3', 'h4', 'h5', 'h6'
    if element.name not in hierarchy:
        raise ValueError(f'Element with {element.name} (expected one of {hierarchy})')

    same_and_higher = hierarchy[: hierarchy.index(element.name) + 1]
    return get_siblings_until(element, same_and_higher)


def get_siblings_until(
    element: PageElement, until: Union[str, Container[str]]
) -> List[PageElement]:
    """
    Return a list of sibling elements until the next occurrence of a certain tag name (or
    list/tuple/etc. of tag names).

    The `element` itself is included, and the found occurrence of `until` is excluded.
    """
    break_names = [until] if isinstance(until, str) else until
    found = [element]
    for sibling in element.next_siblings:
        if sibling.name in break_names:
            break
        else:
            found.append(sibling)
    return found


def standardize_spaces(s: str) -> str:
    """
    Replace non-breaking space U+00a0 with a space

    >>> standardize_spaces('This\u00a0 \u00a0 has some\u00a0nbsps')
    'This    has some nbsps'
    """
    return s.replace('\u00a0', ' ')


def render_list(strlist: Iterable[str], sep='\n', prefix='- ') -> str:
    """
    Take a list like ['foo', 'bar'] and render it as a multiline string like:
    - foo
    - bar

    >>> print(render_list(['foo', 'bar']))
    - foo
    - bar

    >>> render_list(['world', 'again'], sep='||', prefix='hello')
    'helloworld||helloagain'
    """
    return sep.join(f'{prefix}{s}' for s in strlist)



import argparse

import rich.console

from . import __version__
from .core import WordInfo, WiktionaryPageNotFound, LanguageEntryNotFound


def main(args):
    parser = argparse.ArgumentParser(
        prog='palabras',
        description='Look up a word',
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=50),
    )
    parser.add_argument('word', metavar='<word>', type=str, help='A word to look up')
    parser.add_argument(
        '-V', '--version', action='version', version=f'%(prog)s {__version__}'
    )
    parser.add_argument(
        '-r',
        '--revision',
        type=int,
        metavar='<n>',
        help='Wiktionary revision ID (from permalink)',
    )
    parser.add_argument(
        '--compact',
        action='store_true',
        help='List definitions for all parts of speech together',
    )
    parser.add_argument('--json', action='store_true', help='Output as JSON')
    args = parser.parse_args(args)

    console = rich.console.Console()

    try:
        word_info = WordInfo.from_search(args.word, revision=args.revision)
        output = parse(word_info, compact=args.compact, json=args.json)
        console.print(output, crop=False, overflow='ignore')
        return 0
    except (WiktionaryPageNotFound, LanguageEntryNotFound) as exc:
        print(exc)
        return 1


def parse(word_info: WordInfo, compact: bool, json: bool) -> str:
    if json:
        return word_info.json_output()
    elif compact:
        return word_info.compact_definition_output()
    else:
        return word_info.definition_output()
